#[allow(dead_code)]
pub static VERSION: &str = "1.5.2";
pub static PATH_ID: &str = "/com/ranfdev/DistroShelf";
pub static GETTEXT_PACKAGE: &str = "distroshelf";
pub static LOCALEDIR: &str = "/usr/share/locale";
pub static PKGDATADIR: &str = "/usr/share/distroshelf";
#[allow(dead_code)]
pub static GRESOURCE_PATH: &str = "/build/distroshelf/src/build/data/distroshelf.gresource";
